import numpy as np
import pickle
from copy import deepcopy
import collections
from bstset import Set
from tqdm import tqdm
from typing import List, Dict

'''
4 strategies in this module:

greedy
anti_greedy
greedy_v2
degree_weighted
hard_greedy

Plug in 1) adjacency_list 2) budgets
apart from other algorithm specific args, which have defaults
'''

def greedy(adjacency_list: List[List[int]], budgets: List[int], reverse: bool = True) -> Dict[int, List[bool]]:
    '''
    Picks larger degree nodes first
    '''
    degree_vertex_list = [(len(vertex_list), vertex) for vertex, vertex_list in enumerate(adjacency_list)]
    degree_vertex_list.sort(reverse=reverse)
    result_dict = {}
    taken_mask = [False for _ in adjacency_list]
    if 0 in budgets:
        result_dict[0] = deepcopy(taken_mask)
    for budget, (degree, vertex) in enumerate(degree_vertex_list):
        taken_mask[vertex] = True
        if budget + 1 in budgets:
            result_dict[budget + 1] = deepcopy(taken_mask)
    return result_dict


def anti_greedy(adjacency_list: List[List[int]], budgets: List[int]):
    '''
    Picks lower degree nodes first
    '''
    return greedy(
        adjacency_list=adjacency_list,
        budgets=budgets,
        reverse=False
    )


def hard_greedy(adjacency_list: List[List[int]], budgets: List[int]) \
    -> Dict[int, List[bool]]:
    '''
    Picks high degree vertices, and removes their
    neighbours from the dataset and they remain
    unselected from then onwards. This means
    neighbouring vertices will never be selected
    by this strategy.
    '''
    degree_vertex_list = [(len(vertex_list), vertex) for vertex, vertex_list in enumerate(adjacency_list)]
    degree_vertex_list.sort(reverse=True)
    result_dict = {}
    taken_mask = [False for _ in adjacency_list]
    dominated_mask = deepcopy(taken_mask)
    if 0 in budgets:
        result_dict[0] = deepcopy(taken_mask)
    budget = 0
    for _, (degree, vertex) in enumerate(degree_vertex_list):
        if dominated_mask[vertex]:
            continue
        taken_mask[vertex], dominated_mask[vertex] = True, True
        for neighbour in adjacency_list[vertex]:
            dominated_mask[neighbour] = True
        if budget + 1 in budgets:
            result_dict[budget + 1] = deepcopy(taken_mask)
        budget += 1
    return result_dict


def greedy_v2(
    adjacency_list: List[List[int]], 
    budgets: List[int], 
    desc: str = None) -> Dict[int, List[bool]]:
    '''
    Picks lower degree nodes first,
    while updating priorities for nodes
    that no longer dominate neighbour nodes.
    '''
    if len(adjacency_list) == 0:
        return {k: [] for k in budgets}

    node_set = Set()
    scores = [len(i) + 1 for i in adjacency_list]
    for node, score in enumerate(scores):
        node_set.push((score, node))
    taken = [False for _ in adjacency_list]
    dominated = [False for _ in adjacency_list]
    result_dict = {}

    if 0 in budgets:
        result_dict[0] = deepcopy(taken)
    counter = 0
    # start computing dominating sets
    for _ in tqdm(range(min(max(budgets), len(adjacency_list))), desc=desc):
        score, node = node_set.max_item()
        node_set.pop((score, node))
        taken[node] = True

        # now need to update neighbours
        for neighbour in adjacency_list[node]:
            if not taken[neighbour]:
                node_set.pop((scores[neighbour], neighbour))
                scores[neighbour] -= (0 if dominated[node] else 1) + (0 if dominated[neighbour] else 1)
                node_set.push((scores[neighbour], neighbour))
            if not dominated[neighbour]:
                for next_neighbour in adjacency_list[neighbour]:
                    if not taken[next_neighbour] and next_neighbour != node:
                        node_set.pop((scores[next_neighbour], next_neighbour))
                        scores[next_neighbour] -= 1
                        node_set.push((scores[next_neighbour], next_neighbour))

        for neighbour in adjacency_list[node]:
            dominated[neighbour] = True

        dominated[node] = True

        counter += 1
        if counter in budgets:
            result_dict[counter] = deepcopy(taken)
    return result_dict


def _get_ranked_vertices(adjacency_list: List[List[int]], importance_function) -> List[int]:
    importance = [0.0 for adj in adjacency_list]
    for vertex, adj in enumerate(adjacency_list):
        adj_degrees = [len(adjacency_list[neighbour]) for neighbour in adj]
        for adj_degree in adj_degrees:
            importance[vertex] += importance_function(adj_degree)
    vertices = map(lambda x: (x[1], x[0]), enumerate(importance))
    vertices = sorted(vertices, reverse=True)
    vertices = map(lambda x: x[1], vertices)
    return list(vertices)


def _record_to_checkpoints(ranked_vertices: List[int], budgets: List[int]) -> Dict[int, List[bool]]:
    taken_mask = [False for _ in ranked_vertices]
    result_dict = {}
    if 0 in budgets:
        result_dict[0] = deepcopy(taken_mask)
    for i, vertex in enumerate(ranked_vertices):
        taken_mask[vertex] = True
        if i + 1 in budgets:
            result_dict[i + 1] = deepcopy(taken_mask)
    return result_dict


def degree_weighted(adjacency_list: List[List[int]], budgets: List[int], batch_size: int = 128) -> Dict[int, List[bool]]:
    '''
    Degree weighted
    '''
    ranked_vertices = _get_ranked_vertices(adjacency_list, lambda x: 1.0/(1.0+x))
    result_dict = _record_to_checkpoints(ranked_vertices, budgets)
    return result_dict

def degree_range_random(degree_range, adjacency_list, budgets):
    degrees = [(len(adj), i) for i, adj in enumerate(adjacency_list)]
    candidate_indices = [index for (degree, index) in degrees \
        if degree in range(*degree_range)]
    result = {}
    for budget in budgets:
        chosen_indices = np.random.choice(candidate_indices, size=budget, replace=False)
        result[budget] = [index in chosen_indices for index, _ in enumerate(adjacency_list)]
    return result

STRATEGIES = {
    'greedy': greedy,
    'anti_greedy': anti_greedy,
    'greedy_v2': greedy_v2,
    'degree_weighted': degree_weighted,
    'hard_greedy': hard_greedy,
    'degree_range_random': degree_range_random,
}
